#include "param.h"

float4   cnt    : register(c0);
sampler  t0     : register(s0);

struct map_info {
  float t;
  int id;
};


float3 rotate_pos(float3 d, float rot) {
  float3x3 mtx = {
    cos(rot),  sin(rot), 0,
    sin(rot), -cos(rot), 0,
    0,       0,      1
  };
  return mul(mtx, d);
}

//-------------------------------------------------------------
// distance functions
//
// http://www.iquilezles.org/www/articles/distfunctions/distfunctions.htm
//
//-------------------------------------------------------------
float disp(float3 p, float w)      { return sin(w * p.x) * sin(w * p.y) * sin(w * p.z); }
float disp2(float3 p, float w)     { return sin(w * p.x * 0.79) * sin(w * p.y * 0.56) * sin(w * p.z * 0.91); }
float plane(float3 p, float4 n)    { return dot(p, normalize(n.xyz))    + n.w; }
float sphere(float3 p, float s)    { return length(p)         - s;             }
float cyl(float3 p, float3 s)      { return length(p.xz)      - s.xz;          }
float box(float3 p, float3 s)      { return length(max(abs(p) - s.xz,  0));    }
float ibox(float3 p, float3 scale) { return length(max(abs(p) - scale, 0).xz) - scale.xz * 0.2;}


//-------------------------------------------------------------
// map
//-------------------------------------------------------------
map_info map(float3 p) {
  map_info mi;
  mi.t  = 3000;
  mi.id = -1;
  float e = 0;
  
  float pg1 = 0;
  float pg2 = 0;
  float3 tp = 0;
  pg1 += disp(p, 3)  * 0.2;
  //pg1 += disp(p, 5)  * 0.05;
  //pg1 += disp(p, 10) * 0.003;
  
  //sky
  e = plane(p, float4( 0,   1, 0, 1));
  if(e < mi.t)  { mi.t = e; mi.id = 2; }
  
  e = plane(p, float4( 0,  -1, 0, 1));
  if(e < mi.t)  { mi.t = e; mi.id = 2; }
  
  //tetu
  tp = p + pg1;
  tp.x += sin(p.z * 0.3) * 1.3;
  tp.y += cos(p.z * 0.2) * 1.2;
  e = cyl( abs(tp.yzx % 10) - 5, 5.3);
  if(-e > mi.t) { mi.t = -e; mi.id = 3; }
  
  tp = p + pg1;
  e = ibox(abs(tp.xyz % 0.5) - 0.25, 0.2);
  if(-e > mi.t) { mi.t = -e; mi.id = 3; }
  
  tp = p + 0.5;
  e = ibox(abs(tp.xyz % 0.1) - 0.05, 0.025);
  if(-e > mi.t) { mi.t = -e; mi.id = 1; }
  
  return mi;
}

//-------------------------------------------------------------
// shadow
//
// http://www.iquilezles.org/www/articles/rmshadows/rmshadows.htm
//
//-------------------------------------------------------------
float shadow(float3 eye, float3 dir, float mint, float maxt, float k) {
  //k is sofff
  float res = 1;
  for(float t = mint; t < maxt;) {
    float3 temp = eye + dir * t;
    map_info g = map(temp);
    res = min(res, k * g.t / t);
    t += g.t * 0.75;
  }
  return res;
}

//-------------------------------------------------------------
// raycast
//-------------------------------------------------------------
map_info raycast(float3 eye, float3 dir) {
  map_info mi;
  float t = 0;
  for(int i = 0; i < 100; i++) {
    float3 temp = eye + dir * t;
    mi = map(temp);
    //if(mi.t < 0.001) break;
    t += mi.t * 0.85;
  }
  mi.t = t;
  return mi;
}

//-------------------------------------------------------------
// getnormal
//-------------------------------------------------------------
float3 getnormal(float3 ip) {
  float  iit    = map(ip).t;
  float3 h      = float3(0.01, 0, 0);
  return normalize(float3(
    iit - map(ip + h.xyy).t,
    iit - map(ip + h.yxy).t,
    iit - map(ip + h.yyx).t));
}

//-------------------------------------------------------------
// ps_main
//-------------------------------------------------------------
float4 ps_main(float2 vp : vpos) : color {
  float2 uv   = -1 + 2 * vp / (float2(ScreenX, ScreenY));
  if(abs(uv.y) > 0.75) return 0;
  uv.y = -uv.y;
  float time  = cnt.x;
  //rotate
  float3 dir = normalize(float3(uv.x * 1.25, uv.y, 1));
  dir = rotate_pos(dir, cnt.x * 0.1);
  dir = -dir.yzx;
  dir = rotate_pos(dir, cnt.x * 0.2);

  //cast
  float3   eye = float3(cnt.x * 2, 0, 3 * cnt.x + sin(cnt.x * 0.2) * 1.3);
  map_info  mi = raycast(eye, dir);
  float      t = mi.t;
  int       id = mi.id;
  float3   hit = eye + dir * (t - 0.001);

  //get color
  float3 pcolor = float3(1, 1, 1);    //sky
  if(id == 0) pcolor = float3(0.1, 0.3, 0.7);
  if(id == 1) pcolor = float3(0.7, 0.4, 0.3);
  if(id == 2) pcolor = float3(0.5, 0.7, 1.0);
  if(id == 5) pcolor = float3(0.5, 0.5, 0.5);

  //shadow
  float S       = 1;
  
  //calc normal
  float3 LD       = normalize(float3(0.3,  1,  1));
  float3 SunColor = float3(.9, .6, .4);
  float3 N        = getnormal(hit);
  float3 D        = max(dot(LD, N), 0.1);
  //float3 Sun      = pow(dot(dir, -LD), 256) * SunColor;
  float3 fog      = t * 0.012;

  if(id == 2) {
    fog = t * 0.01;
    D = 1;
    float k = 0.23333;
    pcolor += disp2(hit + cnt.x * .4, k) * k;
    k *= 0.7;
    pcolor += disp2(hit + cnt.x * .2, k) * k;
    k *= 1.7;
    pcolor += disp2(hit + cnt.x * .3, k) * k;
  }
  //float3 col = fog + pcolor * D * S + Sun;
  float3 col = fog + pcolor * D * S;
  
  return float4(col, t);
}



float4 ps_main_f(float2 vp : vpos, float2 uv : texcoord0) : color {
  float2 k = float2(0.001, 0);
  float4 c = tex2D(t0, uv);
  c += tex2D(t0, uv +  k.xy);
  c += tex2D(t0, uv + -k.xy);
  c += tex2D(t0, uv +  k.yx);
  c += tex2D(t0, uv + -k.yx);
  return c / 5 - dot(uv, uv) * 0.25;
}

